local ai = load_ai_personality("commonguy.inc")
ai.TUNABLE.TURN_ORDER_VALUE = 2
ai.TUNABLE.TAKE_CASTLE_THRESHOLD_001 = 1

local MY_PRESTIGE_VALUES = {
	statue = 10,
	weaver = 12,
	library = 10,
	hotel = 16,
	granary = 10,
	theater = 17,
	college = 17,
	monument = 19,
	cathedral = 25,
	}

ai.TUNABLE.BUILDING_VALUES = {
  --
  --wooden buildings that prestigeguy might want to build
  --
	lawyer = 10,          --needed to build residence
	mason = 8,            --needed to build architect
	["wood-2s"] = 3,      --offers a way to get stone
	["other-wood"] = 2,
  --
  --stone buildings that prestigeguy might want to build
  --
	architect = 12,       --needed to build prestige
	["architect(2)"] = 6,
	["stone-ssc"] = 4,
	bank = 7,             --offers a way to get gold
	alchemist = 7,        --offers a way to get gold
	["other-stone"] = 3,
	}

local function can_build_building(bldgref, resources)
	if is_built(bldgref) then
		return false
	end

	if bldgref.cost == "1w,1_" then
		return resources.wood >= 1
			and
			(resources.food + resources.wood + resources.cloth + resources.stone - 1) >= 1
	end

	return check_cost_r(resources, bldgref.cost)
end

local function debug(msg)
	if ai.VERBOSE then
		print(msg)
	end
end

function ai.select_lawyer_move()
	local best_move = nil
	local best_score = -math.huge

	for i,move in ipairs(all_available_moves()) do

		local my_score = 0
		local position = string.match(move, "^build residence 1,(%d+) ")
		if position then

			position = tonumber(position)
			local bldg = state.buildings[position]
			assert(bldg)

			if bldg.ref == CARPENTER1 then
				my_score = 6
			elseif bldg.ref == MARKETPLACE1
				or bldg.ref.id == "neutral-1w"
			then
				my_score = 5
			elseif bldg.ref.id == "neutral-1f/w"
				or bldg.ref.id == "neutral-1c/f"
			then
				my_score = 3
			end

			if not bldg.owner then
				--prefer unowned buildings to own buildings
				my_score = my_score + 1
			end
		end

		if my_score > best_score then
			best_score = my_score
			best_move = move
		end
	end

	return best_move
end

function ai.select_architect_move()
	local best_move = "noaction"
	local best_score = 0

	for i,move in ipairs(all_available_moves()) do

		local my_score = 0
		local bldg_idx = string.match(move, "^build prestige (%d+),")
		if bldg_idx then

			bldg_idx = tonumber(bldg_idx)
			local bldgref = PRESTIGE_BUILDINGS[bldg_idx]
			local v = MY_PRESTIGE_VALUES[bldgref.id]

			if v and v > best_score then
				best_score = v
				best_move = move
			end
		end
	end

	return best_move
end

function ai.value_of_building(bldgref)

	local id = bldgref.id
	if id == "architect" then
		if state.unbuilt["architect"] == 0 then
			--another architect built
			id = "architect(2)"
		end
	end

	local v = ai.TUNABLE.BUILDING_VALUES[bldgref.id]
	if v then

		return v

	elseif bldgref.btype == "wood" then

		return ai.TUNABLE.BUILDING_VALUES["other-wood"]

	elseif bldgref.btype == "stone" then

		return ai.TUNABLE.BUILDING_VALUES["other-stone"]

	else

		error("value_of_building called on a non-wood/non-stone building")
	end

end

local function pick_payment(cost, r)
	if not r then
		local whoami = current_turn()
		local me = state.players[whoami]
		r = me.resources
	end

	if cost == "1w,1_" then
		if r.food >= 1 then
			return "1w,1f"
		elseif r.wood >= 2 then
			return "2w"
		elseif r.cloth >= 1 then
			return "1w,1c"
		elseif r.stone >= 1 then
			return "1w,1s"
		else
			error("can't pay")
		end
	end

	return cost
end

local SUPER_select_builder_move = ai.select_builder_move
function ai.select_builder_move()
	local whoami = current_turn()
	local me = state.players[whoami]
	local bldgref = state.buildings[state.active_building].ref

	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		--building a wooden building

		local best_move = "noaction"
		local best_score = 0

		for i,bref in ipairs(WOOD_BUILDINGS) do
			if can_build_building(bref, me.resources) then
				local my_score = ai.value_of_building(bref)
				if my_score > best_score then
					best_score = my_score
					best_move = string.format("build %s %s na", bref.id, pick_payment(bref.cost))
				end
			end
		end
		return best_move

	elseif bldgref == MASON then
		--building a stone building

		local best_move = "noaction"
		local best_score = 0

		for i,bref in ipairs(STONE_BUILDINGS) do
			if can_build_building(bref, me.resources) then
				local my_score = ai.value_of_building(bref)
				if my_score > best_score then
					best_score = my_score
					best_move = string.format("build %s %s na", bref.id, bref.cost)
				end
			end
		end
		return best_move

	elseif bldgref == LAWYER then

		return ai.select_lawyer_move()

	elseif bldgref == ARCHITECT1 or bldgref == ARCHITECT2 then

		return ai.select_architect_move()

	end

	--shouldn't reach here
	assert(false)
end

function ai.select_worker_action()
	local bldgref = state.buildings[state.active_building].ref

	if state.turn == TURN_OWNER_ACTION then
		local s,m = ai:pick_building_owner_action(bldgref)
		return m
	end

	local s,m = ai:pick_building_pickup_action(bldgref)
	if s > 0 then
		return m
	end

	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

local function logisticfn(t)
	return 1 / (1 + math.exp(-t))
end

--return an array of building refs
--
local function what_buildings_can_i_build(btype, resources)
	local building_set
	if btype == "wood" then
		building_set = WOOD_BUILDINGS
	elseif btype == "stone" then
		building_set = STONE_BUILDINGS
	elseif btype == "prestige" then
		building_set = PRESTIGE_BUILDINGS
	else
		error("invalid btype: "..btype)
	end

	local result = {}
	for i,bldgref in ipairs(building_set) do
		if can_build_building(bldgref, resources) then
			table.insert(result, bldgref)
		end
	end
	return result
end

local function value_of_taking_xxx_helper(r, building_set)
	local best_score = -math.huge
	for i,bldgref in ipairs(building_set) do

		if can_build_building(bldgref, r) then

			local bag = resource_bag.parse(pick_payment(bldgref.cost, r))
			bag = resource_bag.scale(bag, -1)
			local cost_value = ai:value_of_resources_delta(bag, r)

			local my_score = ai.value_of_building(bldgref) + cost_value
			if my_score > best_score then
				best_score = my_score
			end
		end
	end

	return best_score
end

local function value_of_taking_carpenter(r)
	return value_of_taking_xxx_helper(r, WOOD_BUILDINGS)
end

local function value_of_taking_mason(r)
	return value_of_taking_xxx_helper(r, STONE_BUILDINGS)
end

local function count_my_residences()
	local whoami = current_turn()
	local count = 0
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		if state.buildings[i].owner == whoami then
			count = count + 1
		end
	end
	return count
end

local function value_of_taking_architect(r)
	local best_value = 0
	for i,b in ipairs(what_buildings_can_i_build("prestige", r)) do
		local v = MY_PRESTIGE_VALUES[b.id]
		if v and v > best_value then
			best_value = v
		end
	end

	local bailiff_steps_left = CASTLE_BAILIFF_THRESHOLDS[3] - state.bailiff
	local n_residences = count_my_residences()

	if best_value >= 20 then
		return 9
	elseif best_value > 0 and n_residences >= bailiff_steps_left / 2 then
		return 9
	else
		return 0
	end
end

local function value_of_taking_lawyer(r)
	local whoami = current_turn()
	local me = state.players[whoami]
	local game_portion = math.min(1, (state.round - 1) / 14)

	if r.money >= 2 and r.cloth >= 1 then
		return 8 - (me.income - 2) * 2 - 4 * game_portion
	else
		return 0
	end
end

local function value_of_taking_bank(r)
	if r.money >= 3 then
		return 5
	else
		return 0
	end
end

local function count_blocks_already_given_this_phase()
	local whoami = current_turn()
	local x0 = state.castle_phase > 1
		and CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase - 1] + 1
		or 1
	local x1 = CASTLE_CONTRIBUTION_THRESHOLDS[state.castle_phase]

	local count = 0
	for i = x0,x1 do
		if state.contributions[i] == whoami then
			count = count + 1
		end
	end
	return count
end

local function value_of_taking_castle(r, existing_occupants)

	--no point in taking castle if there's nothing to contribute
	local num_blocks_to_give = ai.estimate_maximum_contribution(r)
	if num_blocks_to_give < 1 then
		return -2
	end

	local whoami = current_turn()
	local i_would_be_first = existing_occupants[1] == whoami or not existing_occupants[1]
	local num_blocks_already_given_this_phase = count_blocks_already_given_this_phase()
	assert(state.castle_phase <= 3)
	local num_rounds_left_this_phase = (CASTLE_BAILIFF_THRESHOLDS[state.castle_phase] - state.bailiff) * .75

	if num_blocks_already_given_this_phase == 0 then
		if i_would_be_first then
			return 8
		elseif num_rounds_left_this_phase <= 2 then
			return 2 + 2 * state.castle_phase
		elseif num_blocks_to_give >= ai.TUNABLE.TAKE_CASTLE_THRESHOLD_001 then
			--we have enough resources accumulated s.t.
			--we want to take castle even though we probably
			--won't get the favor
			return 2
		else
			--better to keep the resources on hand
			return 0
		end
	elseif num_blocks_already_given_this_phase == 1 then
		if i_would_be_first then
			return 6
		else
			--having two blocks earns a favor,
			--but want to save resources in first phase of game
			if state.castle_phase == 1 then
				return 0
			else
				return 2
			end
		end
	else
		if i_would_be_first then
			return 3 --for the favor?
		else
			return 0
		end
	end
end

local function value_of_taking_joust(r)
	local game_portion = math.min(1, (state.round - 1) / 14)

	if r.money >= 2 and r.cloth >= 1 and is_built(LAWYER) then
		return 5 + 2 * game_portion
	else
		return 0
	end
end

local function new_needed_struct()
	return {
		food = 0,
		wood = 0,
		stone = 0,
		cloth = 0,
		gold = 0,
		money = 0,
		food_cloth = 0,
		wood_stone_cloth = 0,
		}
end

local function adjust_needed(needed, bldgref)
	if bldgref == JOUST_FIELD then
		needed.cloth = needed.cloth + 1
		needed.money = needed.money + 1
	elseif bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		needed.wood = needed.wood + 1
		needed.food_cloth = needed.food_cloth + 1
	elseif bldgref == MASON then
		needed.stone = needed.stone + 1
		needed.food = needed.food + 1
	elseif bldgref == LAWYER then
		needed.cloth = needed.cloth + 1
		needed.money = needed.money + 1
	elseif bldgref == BANK then
		needed.money = needed.money + 2
	elseif bldgref == ARCHITECT1 or bldgref == ARCHITECT2 then
		needed.gold = needed.gold + 1
	elseif bldgref == CASTLE then
		needed.food = needed.food + 1
		needed.wood_stone_cloth = needed.wood_stone_cloth + 2
	end
end

---Adjust a "resources" structure, indicating how many of each resource
--the player expects to have after activating the referenced building.
--
local function adjust_obtained(r, needed, bldgref)
	if bldgref.actions then
		local s,movestr = ai:pick_building_pickup_action(bldgref)
		local rsrc_str = string.match(movestr, "^take (.*)$")
		if rsrc_str then
			local bag = resource_bag.parse(rsrc_str)
			r.food = r.food + bag.food
			r.wood = r.wood + bag.wood
			r.stone = r.stone + bag.stone
			r.cloth = r.cloth + bag.cloth
			r.gold = r.gold + bag.gold
			r.money = r.money + bag.money
			return
		end
	end

	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		--note: for simplicity sake, we assume the
		--carpenter is going to cost us a wood+food
		r.wood = math.max(0, r.wood - 1)
		r.food = math.max(0, r.food - 1)
		needed.wood = needed.wood - 1
	elseif bldgref == MASON then
		--note: for simplicity sake, we assume the
		--mason is going to cost us a stone+food
		r.stone = math.max(0, r.stone - 1)
		r.food = math.max(0, r.food - 1)
		needed.stone = needed.stone - 1
		needed.food = needed.food - 1
	elseif bldgref == LAWYER then
		r.money = r.money - 1
		r.cloth = r.cloth - 1
		needed.money = needed.money - 1
		needed.cloth = needed.cloth - 1
	elseif bldgref == BANK then
		if r.money >= 2 then
			r.money = r.money - 2
			r.gold = r.gold + 1
		end
		needed.money = needed.money - 2
	elseif bldgref == JOUST_FIELD then
		r.money = math.max(0, r.money - 1)
		r.cloth = math.max(0, r.cloth - 1)
		needed.money = needed.money - 1
		needed.cloth = needed.cloth - 1
	elseif bldgref == CASTLE then
		r.food = math.max(0, r.food - 1)
		needed.food = needed.food - 1
		needed.wood_stone_cloth = needed.wood_stone_cloth - 2
	end
end

local function value_of_taking_peddler(r, bldgref)
	local best_move = "noaction"
	local best_score = 0
	if r.money >= 2 then
		local give_bag = resource_bag.new()
		give_bag.money = 2
		local take_bag = resource_bag.new()
		for i,rsrc in ipairs({"food","wood","stone","cloth"}) do
			take_bag[rsrc] = (bldgref == PEDDLER2 and 2 or 1)
			local bag = resource_bag.subtract(take_bag, give_bag)
			local my_score = ai:value_of_resources_delta(bag, r)
			if my_score > best_score then
				best_score = my_score
				best_move = string.format("exchange %s %s",
					resource_bag.as_string(give_bag),
					resource_bag.as_string(take_bag))
			end
			take_bag[rsrc] = 0
		end
	end
	return best_score,best_move
end

local function value_of_taking_merchant_guild(r)
	--TODO
	--determine value of an early move of the provost...
	--if it activates a building at or beyond the bailiff,
	--it's probably worth at least 2.0

	return 0
end

local function value_of_taking_stables(r, existing_occupants)
	local whoami = current_turn()
	if #existing_occupants == 0 then
		local current_rank = position_in_turn_order(whoami)
		local new_rank = 1
		return (current_rank - new_rank) * ai.TUNABLE.TURN_ORDER_VALUE
	else
		return 0
	end
end

function ai.value_of_worker_at(i, r)
	local whoami = current_turn()
	local bldgref = assert(state.buildings[i].ref)

	local my_score = ai:pick_building_pickup_action(bldgref)
	if bldgref == CARPENTER1 or bldgref == CARPENTER2 then
		my_score = value_of_taking_carpenter(r)
	elseif bldgref == MASON then
		my_score = value_of_taking_mason(r)
	elseif bldgref == LAWYER then
		my_score = value_of_taking_lawyer(r)
	elseif bldgref == ARCHITECT1 or bldgref == ARCHITECT2 then
		my_score = value_of_taking_architect(r)
	elseif bldgref == BANK then
		my_score = value_of_taking_bank(r)
	elseif bldgref == PEDDLER1 or bldgref == PEDDLER2 then
		my_score = value_of_taking_peddler(r,bldgref)
	elseif bldgref == MERCHANT_GUILD then
		my_score = value_of_taking_merchant_guild(r)
	elseif bldgref == JOUST_FIELD then
		my_score = value_of_taking_joust(r)
	elseif bldgref == STABLES then
		my_score = value_of_taking_stables(r, state.buildings[i].occupants)
	elseif bldgref == CASTLE then
		my_score = value_of_taking_castle(r, state.buildings[i].occupants)
	end
	if i < CASTLE_POSITION then
		my_score = my_score * logisticfn(state.provost - i)
	end
	return my_score
end

function ai.select_worker_placement()
	local whoami = current_turn()
	local me = state.players[whoami]
	local valid_moves = {}
	for i,m in ipairs(all_available_moves()) do
		valid_moves[m] = true
	end

	--determine what resources will be needed
	local needed = new_needed_struct()
	for i = 1, CASTLE_POSITION do
		if state.buildings[i].occupants[1] == whoami then
			local bldgref = assert(state.buildings[i].ref)
			adjust_needed(needed, bldgref)
		end
	end

	--determine what resources we will have at each building
	--we are going to consider placing a worker
	local r = resource_bag.clone(me.resources)
	local nn = resource_bag.clone(needed)

	--precalculate the value of each worker already placed.
	--when a new worker is considered, we check the successor
	--workers to see if any of their values changed (as a
	--result of activating the newly placed worker).
	--E.g. placing a Joust worker may impact the value of the
	--existing Lawyer or Castle workers, since the Joust takes
	--away a coin and cloth.

	local existing_worker_values = {}
	for i = 1, CASTLE_POSITION do
		if state.buildings[i].occupants[1] == whoami then
			existing_worker_values[i] = ai.value_of_worker_at(i, r)
			adjust_obtained(r, nn, state.buildings[i].ref)
		end
	end

	r = resource_bag.clone(me.resources)

	local best_move = "pass"
	local best_score = .1
	for i = 1, CASTLE_POSITION do
		local bldgref = state.buildings[i].ref
		if state.buildings[i].occupants[1] == whoami then
			assert(bldgref)
			adjust_obtained(r, needed, bldgref)
		end

		local movestr = string.format("claim %d", i)
		if valid_moves[movestr] then
			assert(bldgref)
			local my_score = ai.value_of_worker_at(i, r)
			if my_score > best_score then

				--this move seems good, but does it interfere with building activations after this one?
				local rr = resource_bag.clone(r)
				local nn = resource_bag.clone(needed)
				adjust_obtained(rr, nn, bldgref)
				for j = i+1, CASTLE_POSITION do
					if state.buildings[j].occupants[1] == whoami then
						local delta = ai.value_of_worker_at(j, rr) - existing_worker_values[j]
						if delta ~= 0 then
							debug(string.format("value of %s adjusted by %.1f from placing %s",
								state.buildings[j].ref.id,
								delta,
								bldgref.id))
						end
						my_score = my_score + delta
						adjust_obtained(rr, nn, state.buildings[j].ref)
					end
				end
			end

			local bldg_cost
			do
				local bag = resource_bag.new()
				bag.money = -get_cost_to_place_worker(whoami, state.buildings[i])
				bldg_cost = ai:value_of_resources_delta(bag)
			end

			debug(string.format("move %s is worth %.1f (%.1f)", movestr, my_score, bldg_cost))

			my_score = my_score + bldg_cost
			if my_score > best_score then
				best_move = movestr
				best_score = my_score
			end

		end
	end

	debug("selected " .. best_move)

	return best_move
end

function ai.select_favor_resource(favor_options)

	local best_move = assert(favor_options[1])
	local best_score = -math.huge

	for i,movestr in ipairs(favor_options) do
		local rsrc_str = string.match(movestr, "^favor (.*)$")
		if rsrc_str and not string.match(movestr, "^favor exchange ")
		then
			local bag = resource_bag.parse(rsrc_str)
			local my_score = ai:value_of_resources_delta(bag)
			if my_score > best_score then
				best_score = my_score
				best_move = movestr
			end
		end
	end

	return best_move
end

function ai.select_favor_move()
	local whoami = current_turn()
	local me = state.players[whoami]

	local favor_options = {
		vp = {},
		money = {},
		resources = {},
		builder = {},
		builder_nil = {},
		}
	for i,m in ipairs(all_available_moves()) do
		if string.match(m, "^favor (%d+)vp$") then
			table.insert(favor_options.vp, m)
		elseif string.match(m, "^favor (%d+)%$$") then
			table.insert(favor_options.money, m)
		elseif m == "favor build" then
			table.insert(favor_options.builder_nil, m)
		elseif string.match(m, "^favor wood ")
			or string.match(m, "^favor stone ")
			or string.match(m, "^favor residence ")
			or string.match(m, "^favor prestige ")
		then
			table.insert(favor_options.builder, m)
		else
			table.insert(favor_options.resources, m)
		end
	end

	if me.resources.money < 3 and favor_options.money[1] then
		return favor_options.money[1]
	elseif favor_options.resources[1] then
		return ai.select_favor_resource(favor_options.resources)
	elseif favor_options.builder_nil[1] and me.build_favor == 0 then
		return favor_options.builder_nil[1]
	else
		return (all_available_moves())[1]
	end
end

load_ai_mixin(ai, "resource_appraiser.inc")
load_ai_mixin(ai, "provost_helper.inc")

return ai
