local ai = ai

assert(ai)
assert(ai.value_of_resources)

--estimate how valuable it is to have a building activate
local VALUE_OF_ACTIVATION = 3
--a value >0, the smaller the value the more aggressively we assume
--other players will take the best move
local ALPHA = .2
--generic benefit (per step) of making other players move the provost
local BENEFIT_OF_FORCE = .5

local function get_probabilities(player_id, position_values, cost_for_mvmt)

	local position_movement_probabilities = {}
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		local value_to_move = {}
		local best = -math.huge
		for steps = -3,3 do
			if cost_for_mvmt[steps]
				and position_values[i + steps]
			then
				local my_value = position_values[i + steps] - cost_for_mvmt[steps]
				value_to_move[steps] = my_value
				if my_value > best then
					best = my_value
				end
			end
		end

		local prob_helper = {}
		local sum_helper = 0
		for steps = -3,3 do
			if value_to_move[steps] then
				prob_helper[steps] = 1 / (ALPHA + best - value_to_move[steps])
				sum_helper = sum_helper + prob_helper[steps]
			end
		end
		local probabilities = {}
		for steps = -3,3 do
			if prob_helper[steps] then
				probabilities[steps] = prob_helper[steps] / sum_helper
			--else
			--	probabilities[steps] = 0
			end
		end

		position_movement_probabilities[i] = probabilities

		--debug
		if false then
			local tmp_strings = {}
			for steps = -3,3 do
				if probabilities[steps] then
					table.insert(tmp_strings,
						string.format("%d=%.2f",
							steps,
							probabilities[steps]))
				end
			end
			print(string.format("player %s at %d: probability %s",
				player_id,
				i,
				table.concat(tmp_strings, ", ")))
		end
	end
	return position_movement_probabilities

end

local function dump_move_options(position_values, cost_for_mvmt, probabilities)
	if not ai.VERBOSE then
		return
	end

	print("provost movement options")
	for steps = -3, 3 do
		if cost_for_mvmt[steps] then
			assert(position_values[state.provost + steps])

			local probs = {}
			if probabilities[state.provost+steps] then
				for m,p in pairs(probabilities[state.provost+steps]) do
					if p >= .05 then
						table.insert(probs,
							string.format("%d=%d%%",
							m, p*100
							))
					end
				end
				table.sort(probs)
			end
			local probs_str = table.concat(probs, ",")

	print(string.format("  to move %d: good %.1f cost %.1f (%s)",
			steps,
			position_values[state.provost + steps],
			cost_for_mvmt[steps],
			probs_str))

		end
	end
end

local function pick_best_move(position_values, cost_for_mvmt)
	local best_score = -math.huge
	local best_move = "noaction"

	for steps = -3, 3 do
		if cost_for_mvmt[steps] and position_values[state.provost + steps] then
			local my_score = position_values[state.provost + steps]
				- cost_for_mvmt[steps]
			if my_score > best_score then
				best_score = my_score
				best_move = steps == 0 and "noaction"
					or string.format("provost %d", steps)
			end
		end
	end

	if ai.VERBOSE then
		print("  -->"..best_move)
	end
	return best_move
end

local function generate_cost_for_mvmt(pid)
	local cost_for_mvmt = {}
	do
		local bag = resource_bag.clone(state.players[pid].resources)
		local v = ai.value_of_resources(bag)
		cost_for_mvmt[0] = 0
		for i = 1,3 do
			if bag.money >= 1 then
				bag.money = bag.money - 1
				cost_for_mvmt[i] = v - ai.value_of_resources(bag)
				cost_for_mvmt[-i] = cost_for_mvmt[i]
			end
		end
	end
	return cost_for_mvmt
end

local function combine_probabilities(first_player_probs, rest_players_probs)
	local new_probabilities = {}
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do

		local result = {}
		for j,first_player_v in pairs(first_player_probs[i]) do
			for k,rest_players_v in pairs(rest_players_probs[i+j]) do
				result[j+k] = (result[j+k] or 0) + first_player_v * rest_players_v
			end
		end
		new_probabilities[i] = result
	end

	return new_probabilities
end

local function benefit_of_forcing_mvmt(steps)
	return BENEFIT_OF_FORCE * math.abs(steps)
end

--- Choose a valid move when the merchant guild or bridge is being activated.
-- @return a movestr
function ai.select_provost_move()
	local whoami = current_turn()
	local me = state.players[whoami]
	local is_free = (state.active_building == MERCHANT_GUILD_POSITION)

	local per_player_valuations = {}
	local per_player_valuation_last = {}
	for pid in pairs(state.players) do
		per_player_valuation_last[pid] = 0
	end
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		local bldg = state.buildings[i]
		local occupant = bldg.occupants[1]
		if occupant then
			per_player_valuation_last[occupant] = per_player_valuation_last[occupant] + VALUE_OF_ACTIVATION
		end
		per_player_valuations[i] = clone1(per_player_valuation_last)
	end

	local function is_legal_provost_move(steps)
		if state.provost + steps <= BRIDGE_POSITION then
			return false
		end
		if not is_free then
			local cost = math.abs(steps)
			if cost > me.resources.money then
				return false
			end
		end
		return true
	end

	local function value_of_provost_at(provost_pos, whoami)
		local my_value = 0
		local sum_opponents = 0
		if not per_player_valuations[provost_pos] then
			return 0
		end
		for pid,v in pairs(per_player_valuations[provost_pos]) do
			if pid == whoami then
				my_value = my_value + v
			else
				sum_opponents = sum_opponents + v
			end
		end
--print(string.format("  at %d : %s", provost_pos,
--		table.concat(
--			map(function(x) return string.format("%.1f", x) end, per_player_valuations[provost_pos]),
--			",")))


		return my_value - sum_opponents / (#state.players - 1)
	end

	local bridge_bldg = state.buildings[BRIDGE_POSITION]
	local bridge_turn = #bridge_bldg.occupants
	local bridge_occupant_pid = bridge_bldg.occupants[bridge_turn]

	--Note: to make things easier to discuss, we designate the last
	--bridge occupant as "player Z", the second-to-last bridge occupant
	--as "player Y", and so on.

	local position_values = {}
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		position_values[i] = value_of_provost_at(i, bridge_occupant_pid)
	end

	local cost_for_mvmt = generate_cost_for_mvmt(bridge_occupant_pid)
	if state.turn == bridge_turn then
		dump_move_options(position_values, cost_for_mvmt, {})
		return pick_best_move(position_values, cost_for_mvmt)
	end

	--Calculate the probabilities of player Z moving the provost
	--from particular spots; indexed first by provost position,
	--then by steps to move the provost.

	local position_movement_probabilities = get_probabilities(bridge_occupant_pid, position_values, cost_for_mvmt)

	while bridge_turn >= 1 do

		--second-to-last bridge occupant (player Y)

		bridge_turn = bridge_turn - 1
		if bridge_turn >= 1 then
			bridge_occupant_pid = bridge_bldg.occupants[bridge_turn]
		else
			bridge_occupant_pid = state.buildings[MERCHANT_GUILD_POSITION].occupants[1]
		end

		assert(bridge_occupant_pid, "don't know who I'm moving provost for")

		position_values = {}
		for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
			position_values[i] = value_of_provost_at(i, bridge_occupant_pid)
		end

		--adjust position values using next players' probabilities
		local adjusted_values = {}
		for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do

			local sum = 0
			for steps,p in pairs(position_movement_probabilities[i]) do
				assert(position_values[i + steps], "no value available for position "..(i+steps).." arrived at by mvmt "..steps)
				sum = sum + p * (position_values[i + steps] + benefit_of_forcing_mvmt(steps))
			end

			adjusted_values[i] = sum
		end
		position_values = adjusted_values

		if bridge_turn > 0 then
			cost_for_mvmt = generate_cost_for_mvmt(bridge_occupant_pid)
		else
			cost_for_mvmt = { [-3] = 0, [-2] = 0, [-1] = 0, [0] = 0, [1] = 0, [2] = 0, [3] = 0 }
		end

		if bridge_turn == 0 or (state.turn == bridge_turn and state.active_building == BRIDGE_POSITION) then
			dump_move_options(position_values, cost_for_mvmt, position_movement_probabilities)
			return pick_best_move(position_values, cost_for_mvmt)
		end

		--If we got this far, then we are considering provost movement for
		--the merchant hall or a bridge occupant other than the last or
		--second-to-last bridge occupant.

		--To continue we want to update the following variables:
		-- position_movement_probabilities- should now be the probability
		--     that the provost will move a particular distance with BOTH
		--     player Y *and* player Z having a chance to move it.
		--
	
		position_movement_probabilities = combine_probabilities(
			get_probabilities(bridge_occupant_pid, position_values, cost_for_mvmt),
			position_movement_probabilities
			)

	end --loop through bridge occupants

	return "noaction"
end

return ai
