local quickmove = {}
local fallback_ai = load_ai_personality("planner4.lua")

function quickmove.select_builder_move()
	local moves_array = all_available_moves()
	if #moves_array >= 2 and moves_array[1] == "noaction" then
		return moves_array[2]
	else
		return moves_array[1]
	end
end

function quickmove.select_provost_move()
	local whoami = current_turn()
	local is_free = (state.active_building == MERCHANT_GUILD_POSITION)
	if not is_free then
		return "noaction"
	end

	local last_my_worker = 0
	local last_opponent_worker = 0
	local need_to_save_coins = false
	for i = BRIDGE_POSITION + 1, LAST_EMPTY_POSITION do
		local bldg = state.buildings[i]
		if bldg.occupants[1] == whoami then
			last_my_worker = i
			if bldg.ref.id == "lawyer"
				or bldg.ref.id == "bank"
				or bldg.ref.id == "church"
				or bldg.ref.id == "wood-peddler"
				or bldg.ref == "fixed-peddler" then
				need_to_save_coins = true
			end
		elseif bldg.occupants[1] then
			last_opponent_worker = i
		end
	end
	if last_my_worker > state.bailiff - 3 then
		return "provost 3"
	else
		return "provost -3"
	end
end

local preferred_moves = {}
function quickmove.select_move()
	local ph = current_phase()
	if ph == PHASE_WORKER_PLACEMENT then
		return "pass"
	elseif ph == PHASE_MOVE_PROVOST then
		return quickmove.select_provost_move()
	else
		local whoami = current_turn()
		if state.active_building > 0 and state.active_building ~= BRIDGE_POSITION then
			local bldgref = state.buildings[state.active_building].ref
			if bldgref == LAWYER
				or bldgref == ARCHITECT1
				or bldgref == ARCHITECT2
				or bldgref == CARPENTER1
				or bldgref == CARPENTER2
			then
				return quickmove.select_builder_move()
			end
			local key = string.format("%d,%d", whoami, state.active_building)
			local m = preferred_moves[key]
			if m and check_move(m) then
				return m
			end
			m = fallback_ai.select_move()
			preferred_moves[key] = m
			return m
		else
			return fallback_ai.select_move()
		end
	end
end

return quickmove
