local ai = ai

assert(ai)

--In this "mix-in", we try to correctly quantify the value of having
--unspent resources. This includes money, food, wood, stone, cloth,
--and gold. The number returned is, theoretically, the number of victory
--points these unspent resources will translate to.

local function my_is_built(bldgref)
	if not ai.context.is_built_by_bref then
		local tmp = {}
		for i,bref in ipairs(WOOD_BUILDINGS) do
			tmp[bref] = is_built(bref)
		end
		for i,bref in ipairs(STONE_BUILDINGS) do
			tmp[bref] = is_built(bref)
		end
		for i,bref in ipairs(PRESTIGE_BUILDINGS) do
			tmp[bref] = is_built(bref)
		end
		ai.context.is_built_by_bref = tmp
	end
	return ai.context.is_built_by_bref[bldgref]
end
local is_built = my_is_built

local function can_build_building(bldgref, resources)
	if is_built(bldgref) then
		return false
	end

	if bldgref.cost == "1w,1_" then
		return resources.wood >= 1
			and
			(resources.food + resources.wood + resources.cloth + resources.stone - 1) >= 1
	end

	return check_cost_r(resources, bldgref.cost)
end

local future_workers
local used_workers

local function resource_appraiser3(r)

	return .333 * (r.food + r.wood + r.stone + r.cloth)
		+ 3 * r.gold
		+ .25 * r.money
end

local function resource_appraiser2(r)

	local best_value = resource_appraiser3(r)

	local fw = future_workers["church"] - used_workers["church"]
	if fw > 0 then

		used_workers["church"] = used_workers["church"] + 1
		if r.money >= 4 then

			r.money = r.money - 4
			local value_4 = 5 + resource_appraiser2(r)
			r.money = r.money + 4

			if value_4 > best_value then
				best_value = value_4
			end
		end
		if r.money >= 2 then

			r.money = r.money - 2
			local value_2 = 3 + resource_appraiser2(r)
			r.money = r.money + 2

			if value_2 > best_value then
				best_value = value_2
			end
		end
	end

	return best_value
end

local function resource_appraiser1(r)
	
	local best_value = resource_appraiser2(r)

	local fw = future_workers["jeweler"] - used_workers["jeweler"]
	if fw > 0 then

		used_workers["jeweler"] = used_workers["jeweler"] + 1
		if r.gold >= 1 then

			r.gold = r.gold - 1
			local value_1g = 5 + resource_appraiser1(r)
			r.gold = r.gold + 1

			if value_1g > best_value then
				best_value = value_1g
			end
		end
		if r.gold >= 2 then

			r.gold = r.gold - 2
			local value_2g = 9 + resource_appraiser1(r)
			r.gold = r.gold - 2

			if value_2g > best_value then
				best_value = value_2g
			end
		end
	end

	return best_value
end

local function init_workers_struct()
	return {
		church = 0,
		jeweler = 0,
		}
end

--@param r a resources struct
--
function ai.value_of_resources(r)
	ai.context.rounds_left = (ai.context.rounds_left or 1)
	if ai.context.future_workers then
		future_workers = ai.context.future_workers
	else
		future_workers = init_workers_struct()

		local whoami = current_turn()
		for i = (ai.context.position or 0) + 1, CASTLE_POSITION do
			local bldg = state.buildings[i]
			if bldg.ref and bldg.occupants[1] == whoami then
				future_workers[bldg.ref.id] = 1
			elseif bldg.ref and not bldg.occupants[1] then
				future_workers[bldg.ref.id] = .5
			elseif bldg.ref then
				--opponent has this building this round,
				--maybe I can get it next round?
				future_workers[bldg.ref.id] = math.max(0,(ai.context.rounds_left - 1) / 2)
			end
		end
		ai.context.future_workers = future_workers
	end

	used_workers = init_workers_struct()

	local v = resource_appraiser1(resource_bag.clone(r))
	print(string.format("value of %s is %.1f", resource_bag.as_string(r), v))
	return v
end


return ai
