------------------------------------------------------------------------
-- util.lua
--
-- Contains functions not used by the GUI or AI, but maybe by some of
-- the command-line tools written in Lua.
--
------------------------------------------------------------------------

function show_game_state()
	print(string.format("Phase %d of round %d", state.phase, state.round))
	print("Players:")
	for i,player_id in ipairs(state.order) do
		local info = state.players[player_id]
		local tmp_list = {}
		if info.resources.food > 0 then
			table.insert(tmp_list, string.format("%d food", info.resources.food))
		end
		if info.resources.wood > 0 then
			table.insert(tmp_list, string.format("%d wood", info.resources.wood))
		end
		if info.resources.stone > 0 then
			table.insert(tmp_list, string.format("%d stone", info.resources.stone))
		end
		if info.resources.cloth > 0 then
			table.insert(tmp_list, string.format("%d cloth", info.resources.cloth))
		end
		if info.resources.gold > 0 then
			table.insert(tmp_list, string.format("%d gold", info.resources.gold))
		end
		local resources_as_string = table.concat(tmp_list, ", ")
		local workers_as_string = state.phase == PHASE_WORKER_PLACEMENT
				and string.format("; workers: %d%s", info.workers_left, (info.passed and "(passed)" or ""))
				or ""
  		local info_as_string = string.format("score: %d; money: %d; resources: %s%s",
				info.score,
				info.resources.money,
				resources_as_string,
				workers_as_string
				)
		local is_turn = (state.turn ~= TURN_BOARD_ACTION and state.turn ~= 0 and player_id == current_turn()) and "*" or " "
		print(string.format("%s%-7s %s", is_turn, info.info.name, info_as_string))
	end
	print(string.format("%-26s%-20s", "Buildings:", "Occupied by:"))
	local gap = false
	for i,info in ipairs(state.buildings) do
		local name = info.ref ~= nil and info.ref.name or ""
		local flag_provost = i == state.provost and "P" or " "
		local flag_bailiff = i == state.bailiff and "B" or
				i == CASTLE_BAILIFF_THRESHOLDS[1] and "-" or
				i == CASTLE_BAILIFF_THRESHOLDS[2] and "-" or
				i == CASTLE_BAILIFF_THRESHOLDS[3] and "-" or
				" "
		local flag_active = i == state.active_building and ">" or " "
		if info.ref == nil and i > state.bailiff and i > state.provost then
			if not gap then
				print("   ...")
				gap = true
			end
		else
			gap = false
			local is_owned = info.owner and string.format("(%s)", state.players[info.owner].info.name) or ""
			local is_worked = table.concat(map(function(pid) return (pid ~= 0 and state.players[pid].info.name or "") end, info.occupants), ", ")
			print(string.format("%-26s %s",
				string.format("%s%s%s%s%s", flag_bailiff, flag_provost, flag_active, name, is_owned),
				is_worked))
		end
	end

	--
	-- show castle contributions
	--
	local start = 1
	for i = 1,3 do
		local str = ""
		local found = 0
		for j = start, CASTLE_CONTRIBUTION_THRESHOLDS[i] do
			if state.contributions[j] then
				str = str .. string.sub(state.players[state.contributions[j]].info.name, 1, 1)
				found = found + 1
			
			else
				str = str .. "."
			end
		end
		if found > 0 or i <= state.castle_phase then
			print(string.format("     [%s]", str))
		end
		start = CASTLE_CONTRIBUTION_THRESHOLDS[i] + 1
	end
end
